/**
Message-Passing Interface (MPI)-calling 
  Graphics Language User-interface Toolkit (GLUT)
  wrapper.

This code intercepts GLUT calls that set up keyboard 
and mouse listeners, and replaces them with broadcasts and
receives.  This makes it quite easy to run a GLUT application
on a parallel system like a powerwall without changing much code,
but still get excellent performance because rendering is local.

The notion is that *all* GLUT calls are "collective" calls in MPI.
This lets mpiglut in most cases assume that all processors have 
identical GLUT and OpenGL state.  The only place mpiglut gets in the 
way is to translate coordinates into your local window:
       - Before rendering (but not during!)
       - While receiving mouse events
       - While creating, moving, and querying windows

Orion Sky Lawlor, olawlor@acm.org, 2006/09/19 (Public Domain)
*/
#ifndef __UAF_MPIGLUT_H
#define __UAF_MPIGLUT_H

#include <mpi.h>
#include <GL/glut.h>

#define MGAPI /* empty, but possibly used for Windows DLLs later */
#ifndef MGAPIENTRY
#define MGAPIENTRY /* empty, but possibly used for Windows DLLs later */
#endif

#ifdef __cplusplus
extern "C" {
#endif

/** Calls MPI_Init, and then glutInit */
MGAPI void MGAPIENTRY mpiglutInit( int* pargc, char** argv );

/** Handles all network and GLUT requests */
MGAPI void MGAPIENTRY mpiglutMainLoop(void);

/** On the master, broadcast redisplay request to all screens */
MGAPI void MGAPIENTRY mpiglutPostWindowRedisplay( int window );
/** On the master, broadcast redisplay request to all screens */
MGAPI void MGAPIENTRY mpiglutPostRedisplay( void );
/** Optionally synchronize processors before swapping buffers */
MGAPI void MGAPIENTRY mpiglutSwapBuffers( void );

/** Return shift, ctrl, alt state broadcast from the master */
MGAPI int MGAPIENTRY mpiglutGetModifiers(void);

/** We store the user's display function */
MGAPI void MGAPIENTRY mpiglutDisplayFunc( void (* callback)( void ) );
/** Reshape is called with the size of the entire powerwall's combined screen */
MGAPI void MGAPIENTRY mpiglutReshapeFunc( void (* callback)( int w, int h) );

/** We secretly ask OpenGL for only our portion of the viewport */
MGAPI void MGAPIENTRY mpiglViewport(GLint x,GLint y,GLsizei width,GLsizei height);

/** Mouse movement is broadcast from the master */
MGAPI void MGAPIENTRY mpiglutMouseFunc( void (* callback)( int button, int state, int x, int y) );
MGAPI void MGAPIENTRY mpiglutMotionFunc( void (* callback)( int x, int y) );
MGAPI void MGAPIENTRY mpiglutPassiveMotionFunc( void (* callback)( int x, int y) );
MGAPI void MGAPIENTRY mpiglutEntryFunc( void (* callback)( int state) );

/** Keypresses are broadcast from the master */
MGAPI void MGAPIENTRY mpiglutKeyboardFunc( void (* callback)( unsigned char key, int x,int y) );
MGAPI void MGAPIENTRY mpiglutKeyboardUpFunc( void (* callback)( unsigned char key, int x, int y) );
MGAPI void MGAPIENTRY mpiglutSpecialFunc( void (* callback)( int key, int x, int y) );
MGAPI void MGAPIENTRY mpiglutSpecialUpFunc( void (* callback)( int key, int x, int y) );

#ifdef __cplusplus
};
#endif


/* FIXME: 
figure out how to treat windows.
 glutCreateSubWindow, glutReshapeWindow, glutPositionWindow -> should translate coordinates for GLUT
 
Might use glutSetCursor(GLUT_CURSOR_NONE) and glutWarpPointer to extract relative mouse motion, and add up mouse position ourselves.

should probably also handle:
 glutIgnoreKeyRepeat -> send to PE zero (currently OK if called everywhere)
 glutCreateMenu -> stash callback, to allow picking on nonzero PEs

*/

#ifndef MPIGLUT_OVERRIDES_GLUT
/* Route user GLUT calls to mpiglut implementations */
#define glutInit mpiglutInit
#if 0
#define glutMainLoop mpiglutMainLoop

#define glutPostWindowRedisplay mpiglutPostWindowRedisplay
#define glutPostRedisplay mpiglutPostRedisplay
#define glutSwapBuffers mpiglutSwapBuffers

#define glutGetModifiers mpiglutGetModifiers

#define glutDisplayFunc mpiglutDisplayFunc
#define glutReshapeFunc mpiglutReshapeFunc

#define glViewport mpiglViewport

#define glutMouseFunc mpiglutMouseFunc
#define glutMotionFunc mpiglutMotionFunc
#define glutPassiveMotionFunc mpiglutPassiveMotionFunc
#define glutEntryFunc mpiglutEntryFunc

#define glutKeyboardFunc mpiglutKeyboardFunc
#define glutKeyboardUpFunc mpiglutKeyboardUpFunc
#define glutSpecialFunc mpiglutSpecialFunc
#define glutSpecialUpFunc mpiglutSpecialUpFunc
#endif

#endif /* GLUT overrides */

#endif /* this header */
